Require Import List Omega Bool.
Require Import ssreflect ssrbool eqtype ssrnat seq.

Section Arith_and_ListOps.
  
Lemma in_not_nil (n : nat) ls : n \in ls -> exists m ls', ls = m :: ls'.
Proof. case: ls => [|m ls] //. by exists m, ls.
Qed.

Lemma undup_not_nil (n : nat) ls : exists m ls', undup (n :: ls) = m :: ls'.
Proof.  
  elim: ls n => [|m ls IHls] n.
  - by exists n, nil.
  - move: (IHls m) => [k [ls' A]] {IHls}.
    case B: (n \in (m :: ls)) => /=; rewrite B.
    + by exists k, ls'.
    + by exists n, (undup (m :: ls)).
Qed.

Lemma notin_cat (A B : list nat) x : x \notin A -> x \notin B -> x \notin A ++ B.
Proof. move => P Q. move: (negbTE P) (negbTE Q) => R S {P Q}. by rewrite mem_cat R S. Qed.

Lemma cat_notin (A B : list nat) x :
  x \in A ++ B = false -> x \in A = false /\ x \in B = false.
Proof.
  move => P; split; rewrite mem_cat in P;
  by apply orb_false_iff in P as [P Q].
Qed.

Lemma subset_undup (A B : list nat) : {subset A <= B}
  -> {subset (undup A) <= (undup B)}.
Proof.
  move => P x Q. rewrite mem_undup in Q. move: (P x Q) => R {P Q}.
  by rewrite -mem_undup in R.
Qed.

Lemma subset_card (A B : list nat) : {subset A <= B} ->
  size (undup A) <= size (undup B).
Proof.
  move => P. apply subset_undup in P.
  apply uniq_leq_size; last assumption.
  apply undup_uniq.
Qed.

Lemma subseq_card (A B : list nat) : subseq A B ->
  size (undup A) <= size (undup B).
Proof. move => P. apply (subset_card A B (mem_subseq P)). Qed.

Lemma in_subseq (A B : list nat) x :
  x \in B -> subseq A B -> {subset x :: undup A <= undup B}.
Proof.
  move => P R y T. rewrite in_cons in T.
    case: (orb_true_iff (y == x) (y \in undup A)) => Q _;
    move: (Q T) => [S | S] {Q T}.
    - by rewrite (eqP S) mem_undup.
    - rewrite mem_undup in S. move: ((mem_subseq R) y S) => Q {S P}.
      by rewrite mem_undup.
Qed.

Lemma strict_subseq (A B : list nat) : subseq A B ->
  (exists x, x \notin A /\ x \in B) -> size A < size B.
Proof.
  move => P [x [Q R]].
  case U : (A == B).
  - have T := subseq_refl A.
    rewrite {1} (eqP U) in T.
    move : (mem_subseq T R) => {T} T.
    by rewrite T in Q.
  - move : (size_subseq P) => S.
    have T :=  size_subseq_leqif P.
    rewrite -T in U.
    rewrite leq_eqVlt in S.
    by rewrite U in S.
Qed.

Lemma strict_subseq_undup (A B : list nat) : subseq A B ->
  (exists x, x \notin A /\ x \in B) -> size (undup A) < size (undup B).
Proof.
  move => P [x [Q R]]. move: (in_subseq A B x R P) => T.
  have S : size (x :: undup A) <= size (undup B).
  { apply uniq_leq_size ; last by done.
    move => /=. apply andb_true_iff; split.
    by rewrite -mem_undup in Q.
    - apply undup_uniq.
  }
  done.
Qed.

Lemma nil_cat (A B : list nat) : (A ++ B == nil) -> (A == nil).
Proof. by case A, B. Qed.

Lemma cat_nil (A B : list nat) : (A ++ B == nil) -> (B == nil).
Proof. by case A, B. Qed.

End Arith_and_ListOps.


(** Abstract syntax type **)

Inductive term :=
  | Var of nat
  | TT
  | FF
  | And of term & term
  | Or of term & term
  | Not of term
  | Impl of term & term
  | Eq of term & term.

Definition assignment := nat -> bool.


(** decision procedure **)

Fixpoint vars s :=
  match s with
    | Var x => [:: x]
    | TT => nil
    | FF => nil
    | And s t => vars s ++ vars t
    | Or s t => vars s ++ vars t
    | Not s => vars s
    | Impl s t => vars s ++ vars t
    | Eq s t => vars s ++ vars t
  end.

Function nvars s := size (vars s).

Definition subst s x t :=
  (fix subst s :=
     match s with
     | Var y => if x == y then t else Var y
     | TT => TT
     | FF => FF
     | And s t => And (subst s) (subst t)
     | Or s t => Or (subst s) (subst t)
     | Not s => Not (subst s)
     | Impl s t => Impl (subst s) (subst t)
     | Eq s t => Eq (subst s) (subst t)
     end) s.

Fixpoint simplify s : term :=
  match s with
  | And s t => 
    let s' := simplify s in
    let t' := simplify t in
    match s', t' with
    | FF, _ => FF
    | _, FF => FF
    | TT, _ => t'
    | _, TT => s'
    | _, _ => And s' t'
    end
  | Or s t => 
    let s' := simplify s in
    let t' := simplify t in
    match s', t' with
    | TT, _ => TT
    | _, TT => TT
    | FF, _ => t'
    | _, FF => s'
    | _, _ => Or s' t'
    end
  | Not s =>
    let s' := simplify s in
    match s' with
    | TT => FF
    | FF => TT
    | _ => Not s'
    end
  | Impl s t =>
    let s' := simplify s in
    let t' := simplify t in
    match s', t' with
    | FF, _ => TT
    | _, TT => TT
    | TT, _ => t'
    | _, FF => Not s'
    | _, _ => Impl s' t'
    end
  | Eq s t =>
    let s' := simplify s in
    let t' := simplify t in
    match s', t' with
    | TT, _ => t'
    | _, TT => s'
    | FF, _ => Not t'
    | _, FF => Not s'
    | _, _ => Eq s' t'
    end
  | _ => s
  end.

Notation "s °" := (simplify s) (at level 1, no associativity).

Function pick s := ohead (vars s).

Fixpoint taut0 s :=
  match s with
  | Var x => false
  | TT => true
  | FF => false
  | And s t => taut0 s && taut0 t
  | Or s t => taut0 s || taut0 t
  | Not s => ~~ taut0 s
  | Impl s t => taut0 s ==> taut0 t
  | Eq s t => taut0 s == taut0 t
  end.

Fixpoint bounded_taut s n : bool :=
  match n with
  | 0 => false
  | S n => let s := simplify s in
    match pick s with
    | Some x => (bounded_taut (subst s x TT) n) && (bounded_taut (subst s x FF) n)
    | None => taut0 s
    end
  end.

Function taut s := bounded_taut s (nvars s).+1.


(** reflection **)

Lemma nvars0 s : (nvars s == 0) = (vars s == [::]).
Proof. rewrite size_eq0. by case (vars s). Qed.

Lemma vars_subseq s x t: vars t == [::] -> subseq (vars (subst s x t)) (vars s).
Proof.
  elim: s => [n|||s IHs s' IHs'|s IHs s' IHs'|s IHs|s IHs s' IHs'|s IHs s' IHs'] A //=;
                                           try apply (cat_subseq (IHs A) (IHs' A)).
  - case (eqVneq x n) => B.
    + have C : x == n by rewrite B. by rewrite C (eqP A).
    + move: (negbTE B) => C {A B}. have B : vars (Var n) = n :: nil by done.
      rewrite C B {B C}. have A : n == n by done. by rewrite A.
  - apply (IHs A).
Qed.

Lemma subst_away s x t : vars t == [::] -> x \notin vars (subst s x t).
Proof.
  elim: s => [n|||s IHs s' IHs'|s IHs s' IHs'|s IHs|s IHs s' IHs'|s IHs s' IHs'] A //=;
  try (move: (IHs A) => {IHs} IHs); try (move: (IHs' A) => {IHs'} IHs'); try by apply notin_cat.
  - case (eqVneq x n) => [C | C].
    + have D : x == n by rewrite C. by rewrite D (eqP A).
    + have D : vars (Var n) = n :: nil; first by done.
      by rewrite (negbTE C) D /mem /in_mem /= (negbTE C).
  - done.
Qed.

Lemma nvars_And_sum s t: nvars (And s t) = nvars s + nvars t.
Proof. by rewrite /nvars size_cat. Qed.

Lemma nvars_Or_sum s t: nvars (Or s t) = nvars s + nvars t.
Proof. by rewrite /nvars size_cat. Qed.

Lemma nvars_Impl_sum s t: nvars (Impl s t) = nvars s + nvars t.
Proof. by rewrite /nvars size_cat. Qed.

Lemma nvars_Eq_sum s t: nvars (Eq s t) = nvars s + nvars t.
Proof. by rewrite /nvars size_cat. Qed.

Lemma nvars_Not s : nvars (Not s) = nvars s.
Proof. by rewrite /nvars. Qed.
  
Lemma subst_notin s x t : x \in vars s = false -> subst s x t = s.
Proof.
  elim: s => [n|||s IHs s' IHs'|s IHs s' IHs'|s IHs|s IHs s' IHs'|s IHs s' IHs'] /= A //;
    try (move : (cat_notin (vars s) (vars s') x A) => {A} [A B];
    move: (IHs A) (IHs' B) => {IHs IHs' A B} IHs IHs';
    by rewrite IHs IHs').
  - rewrite mem_seq1 in A. by rewrite A.
  - by rewrite (IHs A).
Qed.

Lemma nvars_subst s x t : x \in vars s -> vars t == [::] ->
   nvars (subst s x t) < nvars s.
Proof.
  move => A B. rewrite /nvars.
  have C := vars_subseq s x t B.
  have D := subst_away s x t B.
  apply strict_subseq.
  - apply (vars_subseq s x t B).
  - exists x. done.
Qed.

Lemma subst_closed s x t : vars s == [::] -> subst s x t = s.
Proof. move => P. apply subst_notin. by rewrite (eqP P). Qed.

Lemma simpl_nvars_And_leq s t : nvars (And s t)° <= nvars s° + nvars t°.
Proof.
  case A: s° => [m|||u v|u v|u|u v|u v]; case B: t° => [k|||q r|q r|q|q r|q r] /=;
  try rewrite A => //; try rewrite B => //;
  repeat rewrite nvars_And_sum; 
  try rewrite nvars_And_sum in IHs; 
  try rewrite nvars_And_sum in IHt;
  try (by apply leq_add || by apply leq_addr).
  by apply leq_addr.
Qed.

Lemma simpl_nvars_Or_leq s t : nvars (Or s t)° <= nvars s° + nvars t°.
Proof.
  case A: s° => [m|||u v|u v|u|u v|u v]; case B: t° => [k|||q r|q r|q|q r|q r] /=;
  try rewrite A => //; try rewrite B => //;
  repeat rewrite nvars_Or_sum; 
  try rewrite nvars_Or_sum in IHs; 
  try rewrite nvars_Or_sum in IHt;
  try (by apply leq_add || by apply leq_addr).
  by apply leq_addr.
Qed.

Lemma simpl_nvars_Impl_leq s t : nvars (Impl s t)° <= nvars s° + nvars t°.
Proof.
  case A: s° => [m|||u v|u v|u|u v|u v]; case B: t° => [k|||q r|q r|q|q r|q r] /=;
  try rewrite A => //; try rewrite B => //;
  repeat rewrite nvars_Impl_sum; 
  try rewrite nvars_Impl_sum in IHs; 
  try rewrite nvars_Impl_sum in IHt;
  try (by apply leq_add || by apply leq_addr).
  by rewrite nvars_Not nvars_Impl_sum leq_addr.
Qed.

Lemma simpl_nvars_Eq_leq s t : nvars (Eq s t)° <= nvars s° + nvars t°.
Proof.
  case A: s° => [m|||u v|u v|u|u v|u v]; case B: t° => [k|||q r|q r|q|q r|q r] /=;
  try rewrite A => //; try rewrite B => //;
  repeat rewrite nvars_Eq_sum; 
  try rewrite nvars_Eq_sum in IHs; 
  try rewrite nvars_Eq_sum in IHt;
  try (by apply leq_add || by apply leq_addr).
  - by apply leq_addr.
  - rewrite nvars_Not nvars_Eq_sum. by apply leq_addr.
Qed.

Lemma simpl_nvars_leq s : nvars s° <= nvars s.
Proof.
  elim: s => [n|||s IHs t IHt|s IHs t IHt|s IHs|s IHs t IHt|s IHs t IHt] //.
  - rewrite nvars_And_sum.
    apply (leq_trans (simpl_nvars_And_leq s t)).
      by apply leq_add.
  - rewrite nvars_Or_sum.
    apply (leq_trans (simpl_nvars_Or_leq s t)).
      by apply leq_add.
  - case A: s° => [m|||u v|u v|u|u v|u v] /=; rewrite A; rewrite A in IHs; done.
  - rewrite nvars_Impl_sum.
    apply (leq_trans (simpl_nvars_Impl_leq s t)).
      by apply leq_add.
  - rewrite nvars_Eq_sum.
    apply (leq_trans (simpl_nvars_Eq_leq s t)).
      by apply leq_add.
Qed.

Lemma nvars_simpl_subst n s x t : vars t == [::] -> nvars s <= n ->
  x \in vars s° -> nvars (subst s° x t) < n.
Proof.
  move => A B C.
  have E:= nvars_subst s° x t C A.
  have F:= simpl_nvars_leq s.   
  apply leq_trans with (n := nvars s) => //.
  apply leq_trans with (n := nvars (simplify s)) => //.
Qed.

Definition denote (phi : nat -> bool) :=
  fix denote s :=
    match s with
    | Var x    => phi x
    | TT       => true
    | FF       => false
    | And s t  => denote s && denote t
    | Or s t   => denote s || denote t
    | Not s    => ~~ denote s
    | Impl s t => denote s ==> denote t
    | Eq s t   => denote s == denote t
    end.

Definition of_bool b := if b then TT else FF.

Lemma denote_subst x phi s : 
  denote phi s = denote phi (subst s x (of_bool (phi x))).
Proof.
  elim: s => [y|||s IHs t IHt|s IHs t IHt|s IHs|s IHs t IHt|s IHs t IHt] //=; try by rewrite -IHs -IHt.
  - case: (boolP (x == y)) => // /eqP ->. by case: (phi y).
  - by rewrite -IHs.
Qed.

Lemma substTF phi x s :
  denote phi (subst s x TT) -> denote phi (subst s x FF) -> denote phi s.
Proof. move => H1 H2. rewrite (denote_subst x). by case: (phi x). Qed.

Lemma simpl_denote s phi : denote phi s = denote phi (simplify s).
Proof.
  elim: s => [y|||s IHs t IHt|s IHs t IHt|s IHs|s IHs t IHt|s IHs t IHt] //.
  - change (denote phi s && denote phi t = denote phi (And s t)°).
    rewrite IHs IHt => {IHs IHt}.
    case A: s° => [m|||u v|u v|u|u v|u v] //; case B: t° => [k|||p q|p q|p|p q|p q] //=;
    try rewrite A => {A} //=; try rewrite B => {B} //=;
    try apply andbT; try apply andbF.
  - change (denote phi s || denote phi t = denote phi (Or s t)°).
    rewrite IHs IHt => {IHs IHt}.
    case A: s° => [m|||u v|u v|u|u v|u v] //; case B: t° => [k|||p q|p q|p|p q|p q] //=;
    try rewrite A => {A} //=; try rewrite B => {B} //=;
    try apply orbT; try apply orbF.
  - rewrite /= IHs. by case s°.
  - change (denote phi s ==> denote phi t = denote phi (Impl s t)°).
    rewrite IHs IHt => {IHs IHt}.
    case A: s° => [m|||u v|u v|u|u v|u v] //; case B: t° => [k|||p q|p q|p|p q|p q] //=;
    try rewrite A => {A} //=; try rewrite B => {B} //=;
    try apply implybT; try apply implybF.
  - change (denote phi s == denote phi t = denote phi (Eq s t)°).
    rewrite IHs IHt => {IHs IHt}.
    case A: s° => [m|||u v|u v|u|u v|u v] //; case B: t° => [k|||p q|p q|p|p q|p q] //=;
    try rewrite A => {A} //=; try rewrite B => {B} //=;
    try (by rewrite eqb_id || by rewrite eqbF_neg).
Qed.

Lemma taut0_denote s phi : nvars s == 0 -> taut0 s = denote phi s.
Proof.
  move => A. rewrite nvars0 in A.
  elim: s A => [y|||s IHs t IHt|s IHs t IHt|s IHs|s IHs t IHt|s IHs t IHt] => A //;
  try move: (nil_cat (vars s) (vars t) A) (cat_nil (vars s) (vars t) A) => B C {A} /=;
  try by rewrite (IHs B) (IHt C).
  simpl. by rewrite (IHs A).
Qed.
  
Lemma pick_in x s : pick s = Some x -> x \in vars s.
Proof.
  case P : (vars s) => [|y ys] Q; rewrite /pick P in Q.
  - done.
  - rewrite /ohead in Q. inversion Q. by rewrite mem_head.
Qed.

Lemma pick_none s : pick s = None -> nvars s == 0.
Proof.
  case P : (vars s) => [|y ys] Q; rewrite /pick P in Q.
  - by rewrite /nvars P.
  - rewrite /ohead in Q. inversion Q.
Qed.

Lemma taut_denote s phi: taut s -> denote phi s.
Proof.
  rewrite /taut.
  remember (nvars s).+1 as n.
  have P: n > nvars s by rewrite Heqn.
  move : {Heqn} P.
  elim: n s => [|n IHn] //= s P Q.
  case R: (pick s°) => [x|]; rewrite simpl_denote; rewrite R in Q.
  - apply (substTF phi x); apply IHn.
      + apply nvars_simpl_subst => //. apply pick_in => //.
      + move: (andP Q) => {Q} [Q _]. assumption.
      + apply nvars_simpl_subst => //. apply pick_in => //.
      + move: (andP Q) => {Q} [_ Q]. assumption.
  - rewrite -taut0_denote => //. apply pick_none => //.
Qed.


(** reification **)

(*
The next 5 tactics (inList, addToList, allVars, lookup, translate)
are folllowing roughly chaper 15 of the book
Certified Programming with Dependent Types (Aug 15, 2015)
by Adem Chlipala 
http://adam.chlipala.net/cpdt/
*)
Ltac inList x xs :=
  match xs with
  | [::] => constr:(false)
  | x :: _ => constr:(true)
  | _ :: ?xs' => inList x xs'
  end.

Ltac addToList x xs :=
  let b := inList x xs in
  match b with
  | true => xs
  | false => constr:(x :: xs)
  end.

Ltac allVars xs e :=
  match e with
    | true => xs
    | false => xs
    | ?e1 && ?e2 =>
      let xs := allVars xs e1 in
      allVars xs e2
    | ?e1 || ?e2 =>
      let xs := allVars xs e1 in
      allVars xs e2
    | ~~ ?e1 => allVars xs e1
    | ?e1 ==> ?e2 =>
      let xs := allVars xs e1 in
      allVars xs e2
    | ?e1 == ?e2 =>
      let xs := allVars xs e1 in
      allVars xs e2
    | _ => addToList e xs
  end.

Ltac lookup x xs :=
  match xs with
  | x :: _ => constr:(O)
  | _ :: ?xt =>
    let n := lookup x xt in
    constr:(S n)
  end.

Ltac translate xs b :=
  match b with 
  | true => constr:(TT)
  | false => constr:(FF)
  | ?A && ?B =>
    let s := translate xs A in
    let t := translate xs B in
    constr:(And s t)
  | ?A || ?B =>
    let s := translate xs A in
    let t := translate xs B in
    constr:(Or s t)
  | ~~ ?A =>
    let s := translate xs A in
    constr:(Not s)
  | ?A ==> ?B =>
    let s := translate xs A in
    let t := translate xs B in
    constr:(Impl s t)
  | ?A == ?B =>
    let s := translate xs A in
    let t := translate xs B in
    constr:(Eq s t)
  | _ =>
    let n := lookup b xs in
    constr:(Var n)
  end.


(** preprocessing **)

Lemma eq2bool (a b : bool) : a = b -> a == b.
Proof. move => P; by rewrite P. Qed.

Ltac elim_coq_eq :=
  repeat(
    match goal with
    | [ |- ?G1 = ?G2] => 
      match G2 with
      | true => idtac
      | _ => apply: eqP
      end
    end
  ).

Ltac handle_bool_assum :=
  intros;
  repeat (
  match goal with
    | [ P : ?a = ?b |- _] =>
      match type of a with bool => apply eq2bool in P end
  end
  );
  elim_coq_eq;
  repeat (
    match goal with
    | [ b : is_true _ |- _] => revert b; apply: implyP
    end
  ).

Ltac preproc :=
  handle_bool_assum;
  hnf.


(** final tactic **)

Ltac shandec :=
  preproc;
  match goal with
  | [ |- ?G = true] =>
    let xs := allVars (@nil bool) G in
    let s := translate xs G in
    exact (taut_denote
      s (* s *)
      (fun n => nth false xs n) (* phi *)
      (eq_refl true) (* taut s n *)
    )
  end.